import sys

from difflib import ndiff
from transolution.tm.engine import index
from transolution.tm.lookup import fuzzyLookup
from transolution.path_mod import path
from transolution.tmx import getTMXFromFile,TMXException
port = 6315
version = 'Transolution TM server v0.1'

class TMException(Exception):pass

def getPlainText(nodelist):
    text_nodes = [node.data for node in nodelist if node.nodeType == node.TEXT_NODE]
    return ''.join(text_nodes)

def getKey(data):
    key = []
    retkey = []
    for word in data.split():
        key.append(hash(word))
    key.sort()
    retkey.append(key[0])
    for elem in key:
        if retkey[-1] != elem:
            retkey.append(elem)
    return retkey

from transolution.xmlutils import toxml,parseString

class TMServer(object):
    def __init__(self):
        self.tus = {}

    def version(self):
        return version
    
    def addTU(self,source,target):
        'add source and target xml strings, index them as plain text without tag info'
        source = parseString(source.encode('utf-8')).firstChild
        target = parseString(target.encode('utf-8')).firstChild
        index_text = getPlainText(source.childNodes)
        if index_text.strip():
            print "addTU:",index_text.encode('utf-8')
            #create key
            key = getKey(index_text)
            tmidx.addItem(index.IndexItem(key,(source,target)))
            return True
        else:
            return False

    def lookup(self,lookup):
        'return best match'
        lookup = parseString(lookup.encode('utf-8')).firstChild
        stext = getPlainText(lookup.childNodes)
        print "Lookup:",stext.encode('utf-8')
        if not stext:
            raise TMException('Cant lookup empty string.')

        rkey = getKey(stext)
        tmidx.fuzz = (len(rkey) / 2.5) + 1
        
        res = tmidx.findItem(rkey)
        if not len(res):
            raise TMException('No match found')

        print "Found:",len(res)
        matches = {}
        for tm_source,tm_target in res:
            matches[getPlainText(tm_source.childNodes)] = tm_source,tm_target
        return getBestMatch(stext,matches)

def getBestMatch(slookup,matches):
    for m in fuzzyLookup( slookup, matches.keys(),cutoff=0.7):
        quality,s = m
        tm_source,tm_target = matches[s]
        return quality,toxml(tm_source),toxml(tm_target)
    raise TMException('No good enough match found')

def parseArgs():
    import sys
    from optparse import OptionParser

    usage = "usage: %prog [options] [tmxpath]"

    op = OptionParser(usage=usage)
    op.add_option('-s', '--slang',default = 'en',
                  action='store', type='string', dest='slang',
                  help='TM source language')
    op.add_option('-t', '--tlang',default = 'sv-se',
                  action='store', type='string', dest='tlang',
                  help='TM target language')
    op.add_option('-r', '--recursive',
                  action='store_true', dest='recursive',
                  help='Process files recursive.')
    op.add_option('-f', '--fmask',
                  action='store', type='string', dest='fmask',
                  help='File mask to use when running recursive.')
    op.add_option('-p', '--port',default=port,
                  action='store', type='int', dest='port',
                  help='Port to use.')


    options, args = op.parse_args(sys.argv[1:])
    try:
        pname = path(args[0])
    except IndexError,e:
        options.fnames = []
    else:        
        if options.recursive:
            if not options.fmask:
                op.error("must give filemask option to when recurse")
            else:
                fnames = pname.walkfiles(options.fmask)
        else:
            if options.fmask:
                if not pname.isdir():
                    op.error("must give a directory path when option fmask is given")
                else:
                    fnames = pname.listdir(options.fmask)
            else:
                fnames = [path(f) for f in args]

        options.fnames = fnames
    return options

def importTMX(tmx,tmidx):
##    import codecs
    print "Slang:%s Tlang:%s"%(slang,tlang)
    tm = getTMXFromFile(tmx)
    totents = 0
    skipped = 0
    ents = 0
    for tu in tm.body.tus:
        try:
            key = tu.getKey(slang)
            tu.getKey(tlang)
            val = ( getSeg(tu,slang), getSeg(tu,tlang) )
        except TMXException,e:
            skipped+=1
            continue

        tmidx.addItem(index.IndexItem(key,val))
        ents += 1
        if ents % 100 == 0:
            print ents
    if ents == 0:
        raise SystemExit('No TUs matching Source and Target langs found in TMX.')
    print "Total TU's added:%d skipped:%d"%(ents,skipped)
    del tm
    return tmidx

def getSeg(tu,lang):
    lang = lang.lower()
    for tuv in tu.tuvs:
        if tuv.lang.lower() == lang:
            return tuv.seg.node.cloneNode(True)
    raise TMXException('No TUV with lang code found:%s'%(lang,))

def startServer(slang1,tlang1,fnames,port):
    global slang,tlang,tmidx
    slang = slang1
    tlang = tlang1
    
    tmidx = index.VectorIndex()
    tmidx.nodesize = 40
    tmidx.fuzz = 0

    for tmxfile in fnames:
        print "importing:%s"%(tmxfile,)
        tmidx = importTMX(tmxfile,tmidx)
        print "Done"

    import SimpleXMLRPCServer
    tm = TMServer()
    server = SimpleXMLRPCServer.SimpleXMLRPCServer(("localhost", port))
    server.register_instance(tm)

    #Go into the main listener loop
    print "TM server listening on port",port
    server.serve_forever()
    
def main(options=None):
    if not options:
        # parse commandline
        options = parseArgs()
    # Start the server
    startServer(options.slang,options.tlang,options.fnames,options.port)

if __name__ == '__main__':
##    main()
    tmidx = index.VectorIndex()
    tmidx.nodesize = 40
    tmidx.fuzz = 0

    xml1 = r'''<seg>first <bpt id="1" >xx</bpt>in tag<ept id="1">xx</ept> some more text</seg>'''
    xml2 = r'''<seg>first source <bpt id="1" >&lt;tag&gt;</bpt>in tag<ept id="1">&lt;/tag&gt;</ept> some more text</seg>'''
    xml3 = r'''<seg>first source in tag some more text</seg>'''
    tm = TMServer()
    
    print tm.addTU(xml1,xml2)
    print "--------"
    print tm.lookup(xml3)
