__doc__ = 'main document model'

from transolution.observer import Observable,Observer,Event
from transolution.xliffeditor.general import EditorException
from segments import External,TuSegment,FileInfo,TuInfo
from tree import IndexTree

from tu import TU_UI_UPDATE_EVENT,TU_CHANGE_EVENT

class DOCUMENT_DIRTY_EVENT(Event): pass

class DocumentIndexException(EditorException):pass

class SegUIUpdate(object):
    '''UI update object.
       Holds old text span to update and the new styled text as a list of tuples (style,text)
    '''
    def __init__(self,styledtext,start,end,bg_color,segindex):
        self.styledtext = styledtext
        self.start,self.end = start,end
        self.bg_color = bg_color
        self.segindex = segindex

class DocumentBase(Observable,Observer):
    'Base Document model colletor for External and TuSegment segments'
    def __init__(self):
        Observable.__init__(self)
        Observer.__init__(self)
        self.segs = []
        self.documentFormatter = None
        self.tree = IndexTree()
        self.getTextSpanOfSeg = self.tree.getBounds
        self._initTree = True
        self.changed = False
        
    def initTree(self):
        'initialize the lookup tree upon first use'
        for seg in self.segs:
            seg.format()
        # Build the lookup tree
        self.tree.buildTree(self.segs)
        self._initTree = False

    def setFormatter(self,documentFormatter):
        self.documentFormatter = documentFormatter
        self.documentFormatter.activate()
        self.setDirty()

    def getFormatter(self):
        return self.documentFormatter

    def setDirty(self,*args):
        '*args are segment classes to set to dirty, if no args all are set'
        if (not args):
            for seg in self.segs: seg.dirty = True
        else:
            for seg in self.segs:
                if seg.__class__ in args:
                    seg.dirty = True

        
    def getSegs(self):
        return self.segs

    def getSegCount(self):
        'return total number of segs'
        return len(self.segs)
    
    def addSeg(self,seg):
        'Add a segment to document'
        self.segs.append(seg)
        self.subscribeToObservable(seg, "cbSegUpdate",(TU_UI_UPDATE_EVENT,TU_CHANGE_EVENT))


    def getSegByTextIndex(self,ti):
        if self._initTree:
            self.initTree()
        try:
            seg = self.tree.lookup(ti).seg
        except IndexError,e:
            # ti is probably at last segment
            seg = self.segs[-1]
        start,end = self.getTextSpanOfSeg(seg)
        return seg,start,end

    def getSegUIUpdates(self):
        'get a list of SegUIUpdate with the text span to update and text to update'
        if self._initTree:
            self.initTree()
        getTextSpanOfSeg = self.getTextSpanOfSeg
        for i,seg in enumerate(self.segs):
            seg.format()
            if seg.dirty:
                seg.dirty = False
                start,end = getTextSpanOfSeg(seg)
                yield SegUIUpdate(seg.styledtext,start,end,
                                  bg_color=seg.bg_color,
                                  segindex=i)
                self.tree.updateSeg(seg)


    def cbSegUpdate(self, observable,event,seg):
        'observer interface'
        self.notifyObservers(event,seg)
        if (event is TU_CHANGE_EVENT) and not self.changed:
            self.changed = True
            self.notifyObservers(DOCUMENT_DIRTY_EVENT,seg)



from constraints import unTranslated,unApproved

class DocumentEditor(DocumentBase):
    'extend Base document model with tu segment operations'
    def __init__(self):
        DocumentBase.__init__(self)
        
    def firstTuSegmentSegment(self):
        seg = self.segs[0]
        if not (seg.__class__ is TuSegment):
            seg = self.nextTuSegmentSegment(seg)
        return seg 

    def previousSegment(self,seg):
        i = self.segs.index(seg)-1
        if i < 0: raise DocumentIndexException('At start')
        return self.segs[i]

    def nextSegment(self,seg):
        i = self.segs.index(seg)+1
        if i > len(self.segs)-1: raise DocumentIndexException('At end')
        return self.segs[i]

    def nextTuSegmentSegment(self,seg):
        seg = self.nextSegment(seg)
        while not (seg.__class__ is TuSegment):
            seg = self.nextSegment(seg)
        return seg

    def nextSegmentByConstraint(self,seg,constraint):
        '@constraint is a function that takes a tu and returns true/false'
        seg = self.nextTuSegmentSegment(seg)
        if constraint is None: return seg
        while not constraint(seg):
            seg = self.nextTuSegmentSegment(seg)
        return seg

    def nextUntranslatedSegment(self,seg):
        return self.nextSegmentByConstraint(seg,unTranslated)


    def previousTuSegmentSegment(self,seg):
        seg = self.previousSegment(seg)
        while not (seg.__class__ is TuSegment):
            seg = self.previousSegment(seg)
        return seg

    def previousSegmentByConstraint(self,seg,constraint):
        '@constraint is a function that takes a tu and returns true/false'
        seg = self.previousTuSegmentSegment(seg)
        if constraint is None: return seg
        while not constraint(seg):
            seg = self.previousTuSegmentSegment(seg)
        return seg

    def previousUntranslatedSegment(self,seg):
        return self.previousSegmentByConstraint(seg,unTranslated)

from transolution.xliffeditor.xliffparser_ng import XliffException
   
class ModelDoc(DocumentEditor,Observable):
    '''Document Model
       Adds Observable interface so we can notify GUI to update
    '''
    def __init__(self,xliff=None,formatters={}):
        DocumentEditor.__init__(self)
        Observable.__init__(self)
        self.xliff = xliff
        self.formatters = formatters # document formatters
        self.use_skeleton = True
        self.has_skeleton = False

    def setFormatters(self,formatters):
        'set avaible formatters dict'
        self.formatters = formatters # document formatters

    def parseXliff(self,xliff=None,use_skeleton=True):
        self._initTree = True
        self.has_skeleton = False

        if xliff:
            self.xliff =xliff
        del self.segs
        self.use_skeleton = use_skeleton

        print "Parsing Xliff"
        self.segs = []
        for f in self.xliff.files:
            self.addSeg(FileInfo(f))
            try: f.parseSkeleton()
            except XliffException,e:pass
            else: self.has_skeleton = True

            for seg in f.generateSegments(use_skeleton=use_skeleton):
                self.addSeg(seg)

        print "Xliff Done"

    def delete(self):
        # the tree must be unlinked to free objects
        self.tree.unlink()
        del self.tree
        del self.segs
        del self.xliff
