from transolution.observer import Observable,Observer,Event
from transolution.xmlutils import XmlError,getNodeAttr,setNodeAttr,setCreateNodeAttr,getNodes,getElementsByTagName
from tu_status_wrapper import TuStatusWrapper
from wrappers import NoteWrapper
from state_maps import getStateMap

class TU_OPEN_EVENT(Event): pass
class TU_CLOSE_EVENT(Event): pass
class TU_REMOVE_EVENT(Event): pass
class TU_CHANGE_EVENT(Event): pass
class TU_UI_UPDATE_EVENT(Event): pass

class TuException(Exception):
    pass

class TranslationUnitBase(TuStatusWrapper,Observable):
    'Wraps a Xliff segment to allow Translation unit operations on it'
    def __init__(self,tu_node):
        Observable.__init__(self)

        if tu_node.tagName != 'trans-unit':
            raise ValueError('Node not a trans-unit')
        
        self.changed = True
        self.tu_node = tu_node
        self.open = False
        self._parse()

    def _parse(self):
        'parse tu node to get source & target '
        self.source_node = None
        self.target_node = None
        self.notes = []
        tu = self.tu_node

        self.id = getNodeAttr(tu,'id')
        for node in tu.childNodes:
            if node.nodeType != node.ELEMENT_NODE:
                continue
            
            if node.tagName == 'source':
                self.source_node = node
            elif node.tagName == 'target':
                self.target_node = node
            elif node.tagName == 'note':
                self.notes.append(NoteWrapper(node))

        if not self.source_node:
            raise TuException('no source node found')

    def openForEdit(self):
        'preform an opentu opration and set tu up for edit'
        self.open = True
        # notify observers
        self.notifyObservers(TU_OPEN_EVENT)
        self.notifyObservers(TU_UI_UPDATE_EVENT)


        # save the state
        self._createBackupTarget()
        if self.target_node:
            target_node = self.getTargetNode()
        else:
            self._createEmptyTarget()
            target_node = self.getTargetNode()
        return self.source_node,target_node
    
    def closeForEdit(self,commit,approve=None,state=None,common_state=None):
        'close tu. if not commit restore saved target state'
        if commit:
            self._commitTarget(approve,state=state,common_state=common_state)
        else:
            self._restoreBackupTarget()

        self.open = False
        self.backuptarget = None
        # notify observers
        self.notifyObservers(TU_CLOSE_EVENT)
        self.notifyObservers(TU_UI_UPDATE_EVENT)

    def removeAndCloseForEdit(self):
        'remove target and close tu.'
        self._removeTarget()
        self.open = False
        self.backuptarget = None
        # notify observers
        self.notifyObservers(TU_REMOVE_EVENT)
        self.notifyObservers(TU_UI_UPDATE_EVENT)

        
    def getDocument(self):
        'return nodes document'
        return self.tu_node.ownerDocument

    def copySource(self):
        'Create a target node with same data as source'
        newtarget = self.getDocument().createElementNS(None,'target')
        for node in self.source_node.childNodes:
            newtarget.appendChild(node.cloneNode(True))
        self._insertNewTarget(newtarget)

        self.approved = 'no'

        self.changed = True # set changed flag

    def _createEmptyTarget(self):
        'Create a target with a empty text node'
        doc = self.getDocument()
        newtarget = doc.createElementNS(None,'target')
        newtarget.appendChild(doc.createTextNode(''))
        self._insertNewTarget(newtarget)
        self.approved = 'no'

    def _createBackupTarget(self):
        'Create a backup target node so we can cancel a node edit'
        if self.target_node:
            self.backuptarget = self.target_node.cloneNode(True)
        else:
            self.backuptarget = None

    def _restoreBackupTarget(self):
        'restore the backuped target node on cancel'
        if self.backuptarget:
            self._insertNewTarget(self.backuptarget)
        else:
            self.tu_node.removeChild(self.target_node)
            self._parse()
        self.backuptarget = None

    def _insertNewTarget(self,newtarget):
        'Insert target node at the correct location or replace the excisting'
        if self.target_node:
            self.tu_node.replaceChild(newtarget, self.target_node)
        else:
            nextnode = self.source_node.nextSibling
            if nextnode:
                self.tu_node.insertBefore(newtarget, nextnode)
            else:
                self.tu_node.appendChild(newtarget)
        self._parse()

    def _removeTarget(self):
        'Remove target'
        if self.target_node:
            self.tu_node.removeChild(self.target_node)
            self._parse()
        self.approved = 'no'
        self.changed = True # set changed flag
        if self.backuptarget is not None:
            self.notifyObservers(TU_CHANGE_EVENT)
        self.notifyObservers(TU_UI_UPDATE_EVENT)

    def _commitTarget(self,approve=None,state=None,common_state=None):
        if not self.target_node:
            raise ValueError('not target for TU')
        if not (state is None):
            self.state = state
        elif not (common_state is None):
            # Set the common state accroding to the variant mapper
            self.state = getStateMap()[common_state][0]
            
        # only change approve status if it is a bool type
        if not(approve is None):
            if approve: self.approved = 'yes'
            else:       self.approved = 'no'
        self._parse()
        self.changed = False # unset changed flag
        
        self.notifyObservers(TU_CHANGE_EVENT)
        self.notifyObservers(TU_UI_UPDATE_EVENT)


    def getSourceNode(self):
        return self.source_node

    def getTargetNode(self,createIfNeeded=False):
        if not self.target_node:
            if createIfNeeded:
                self._createEmptyTarget()
            else:
                raise TuException('Tu has no target')
        return self.target_node
    
    def hasTarget(self):
        if self.target_node: return True
        else: return False

    def getStatus(self):
        'Returns the interpreted status using both TU and target nodes'
        return getStatus(self)

from wrappers import AltTransWrapper
class TranslationUnit(TranslationUnitBase):
    'Adds Alt-trans operations'
    def __init__(self,*args,**kwargs):
        TranslationUnitBase.__init__(self,*args,**kwargs)

    def addNote(self):
        'Insert new empty note'
        newnote = self.getDocument().createElementNS(None,'note')
        newnote.appendChild(self.getDocument().createTextNode(''))
        setCreateNodeAttr(newnote,'from','')
        setCreateNodeAttr(newnote,'priority','1')
        setCreateNodeAttr(newnote,'annotates','general')
        self.tu_node.appendChild(newnote)
        self._parse()

##    def _removeTarget(self):
##        'Remove target'
##        if self.target_node:
##            self.tu_node.removeChild(self.target_node)
##            self._parse()
##        self.approved = 'no'
##        self.changed = True # set changed flag


##    def insertNodeAsTarget(self,new_node):
##        'Create a target node with same data alttrans target'
##        newtarget = self.getDocument().createElementNS(None,'target')
##        for node in new_node.childNodes:
##            newtarget.appendChild(node.cloneNode(True))
##        self._insertNewTarget(newtarget)
##        self.approved = 'no'
##        self.changed = True # set changed flag
##
##    def insertAltTransAsTarget(self,alttrans):
##        'Create a target node with same data alttrans target'
##        newtarget = self.getDocument().createElementNS(None,'target')
##        newtarget = alttrans.copyTarget(newtarget)
##
##        self._insertNewTarget(newtarget)
##        self.approved = 'no'
##        self.changed = True # set changed flag

def getStatus(tu):
    'Returns the interpreted status using both TU and target nodes'
    state_map = getStateMap()
    #TODO: implement all possible XLIFF states
    # Dont Translate
    if tu.translate == 'no':  status = 'dont_trans'
    # Approved
    elif tu.approved == 'yes':  status = 'approved'
    # no target
    elif not tu.hasTarget():    status = 'untranslated'
    elif tu.hasTarget():
        state = tu.state
        # final
        if state in state_map['final']:   status = 'final'
        elif state in state_map['signed-off']: status = 'approved'
        # final
        elif state in state_map['needs-translation']:   status = 'needs-translation'
        elif state in state_map['needs-review-translation']:   status = 'needs-translation'
        # new
        elif state in state_map['new']:     status = 'new'
        # translated
        elif state in state_map['translated']: status = 'translated'
        # no state
        elif state is None:      status = 'nostate'
        else: status = 'unknown'
    else:
        raise ValueError('Unknown TU state')

#    print status,tu.open,tu.approved,tu.translate
    return status
