class UndoRedoException(Exception):pass
class UndoException(UndoRedoException):pass
class RedoException(UndoRedoException):pass

class UndoRedo(object):
    'undo redo interface used by the Model interface'
    def __init__(self,*args,**kwargs):
        self.undobuffer = []
        self.redobuffer = []
        
    def insertTextWithUndo(self,ti,text,interactive=True):
        cmd = CmdInsertText(self,ti,text)
        cmd.interactive = interactive
        return self.execute(cmd)

    def deleteTextWithUndo(self,sti,eti,interactive=True,force=False):
        cmd = CmdDeleteText(self,sti,eti,force=force)
        cmd.interactive = interactive
        return self.execute(cmd)

    def cutNodesByTextIndexWithUndo(self,sti,eti,interactive=True):
        cmd = CmdCut(self,sti,eti)
        cmd.interactive = interactive
        return self.execute(cmd)
         

    def pasteNodesAtTextIndexWithUndo(self,sti,nodes,interactive=True):
        cmd = CmdPaste(self,sti,nodes)
        cmd.interactive = interactive
        return self.execute(cmd)
        
    def getState(self):
        'returns a State object'
        raise NotImplementedError('must be implemented in subclasses')
#        return State(formatter = self.formatter,protect_tags = self.protect_tags)
    
    def setState(self,state):
        state.update(self)

    def execute(self,cmd):
        r = cmd.execute()
        self.undobuffer.append(cmd)
        self.redobuffer = []
        return r

    # undo/redo interface
    def undo(self):
        if not self.undobuffer:
            raise UndoException('Nothing to Undo')

        try:
            while 1:
                cmd = self.undobuffer.pop()
                cmd.revert()
                self.redobuffer.append(cmd)
                if cmd.interactive: break
        except IndexError,e:
            pass
        
    def redo(self):
        if not self.redobuffer:
            raise RedoException('Nothing to Redo')

        try:
            while 1:
                cmd = self.redobuffer.pop()
                cmd.execute()
                self.undobuffer.append(cmd)
                if cmd.interactive: break
        except IndexError,e:
            pass

class State(object):
    'object holds state of an xliff object'
    def __init__(self,**kwargs):
        self.dict = kwargs
        
    def xupdate(self, obj,event,msg):
        for key,val in self.dict.items():
            setattr(obj,key,val)


class Command(object):
    def __init__(self,trg,interactive=True):
        self.trg = trg
        self.orgstate = None
        self.interactive = interactive
        
    def execute(self):
        # save the state (formatter etc) at first exec.
        if not self.orgstate:
            self.orgstate = self.trg.getState()

        #save the current state
        state = self.trg.getState()
        # set state to same as first exec time
        self.trg.setState(self.orgstate)

        r = self._execute()

        # restore the current state
        self.trg.setState(state)
        return r

    def revert(self):
        state = self.trg.getState()
        self.trg.setState(self.orgstate)
        self._revert()
        self.trg.setState(state)

    def _execute(self):
        raise NotImplementedError('must be implemented in subclasses')
        
    def _revert(self):
        raise NotImplementedError('must be implemented in subclasses')
        
##class CmdInsertText(Command):
##    def __init__(self,trg,ti,text):
##        Command.__init__(self,trg)
##        self.ti = ti
##        self.text = text
##
##    def _execute(self):
##        self.trg.insertText(self.ti,self.text)
##
##    def _revert(self):
##        self.trg.deleteText(self.ti,self.ti+len(self.text))
##
class CmdInsertText(Command):
    def __init__(self,trg,ti,text):
        Command.__init__(self,trg)
        self.ti = ti
        self.text = text

    def _execute(self):
        self.trg.insertText(self.ti,self.text)

    def _revert(self):
        self.trg.protect_tags = False
        self.trg.deleteText(self.ti,self.ti+len(self.text))

class CmdDeleteText(Command):
    def __init__(self,trg,sti,eti,force=False):
        Command.__init__(self,trg)
        self.sti,self.eti = sti,eti
        self.force = force
    def _execute(self):
        self.deleted = self.trg.deleteText(self.sti,self.eti,self.force)

    def _revert(self):
        self.trg.pasteNodesAtTextIndex(self.sti,self.deleted)

class CmdCut(Command):
    def __init__(self,trg,sti,eti):
        Command.__init__(self,trg)
        self.sti,self.eti = sti,eti

    def _execute(self):
        self.cut = self.trg.cutNodesByTextIndex(self.sti,self.eti)
        return self.cut
    
    def _revert(self):
        self.trg.pasteNodesAtTextIndex(self.sti,self.cut)

class CmdPaste(Command):
    def __init__(self,trg,sti,nodelist):
        Command.__init__(self,trg)
        self.sti,self.nodelist = sti,nodelist

    def _execute(self):
        self.trg.pasteNodesAtTextIndex(self.sti,self.nodelist)

    def _revert(self):
        text = unicode(self.trg.nodelist2string(self.nodelist))
        org = self.trg.protect_tags
        self.trg.protect_tags = False
        self.trg.deleteText(self.sti,self.sti+len(text))
        self.trg.protect_tags = org

class CmdSetFormatter(Command):
    def __init__(self,trg,formatter):
        Command.__init__(self,trg)
        self.formatter = formatter

    def _execute(self):
        self.orgformatter = self.trg.formatter
        self.trg.setFormatter(self.formatter)
        return None
    
    def _revert(self):
        self.trg.setFormatter(self.orgformatter)
