from xliffsegment import XliffSegment,XliffError,Model,FullFormatter,MinimalFormatter
from xml.dom import minidom                                          
import unittest

xml = r'''<seg>first <bpt>&lt;tag&gt;</bpt>in tag<ept>&lt;/tag&gt;</ept> some more text</seg>'''

class XliffTest(unittest.TestCase):
    def setUp(self):
        self.xmldoc = minidom.parseString(xml)
        self.segnode = self.xmldoc.firstChild
        self.nodelist = self.segnode.childNodes

class DocumentTest(XliffTest):
    def testGetDocument(self):
        """Xliff.GetDocument() should return the correct document node"""
        formatter = FullFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        result = xliff.getDocument()
        self.assertEqual(self.xmldoc, result)


class ToTextKnownValues(XliffTest):
        
    def testXml2FullText(self):
        """Test the to full text formatter"""
        result = 'first <tag>in tag</tag> some more text'
        formatter = FullFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        self.assertEqual(str(xliff), result)

    def testXml2MinimalText(self):
        """Test the to minimal text formatter"""
        result = 'first <>in tag</> some more text'
        formatter = MinimalFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        self.assertEqual(str(xliff), result)

class LengthKnownValues(XliffTest):
        
    def testLengthFullText(self):
        """Test get tength using full text formatter"""
        result = 'first <tag>in tag</tag> some more text'
        formatter = FullFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        self.assertEqual(len(xliff), len(result))

    def testLengthMinText(self):
        """Test get tength using minimal text formatter"""
        result = 'first <>in tag</> some more text'
        formatter = MinimalFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        self.assertEqual(len(xliff), len(result))


class CopyNodesByTextIndex(XliffTest):
    def testCopyNodesByTextIndex(self):
        """Copy nodes by text index"""
        tags = ['first ','<tag>','in tag','</tag>',' some more text']
        formatter = FullFormatter()
        org_xliff = XliffSegment(self.segnode,formatter,protect_tags = False)
        for s,e in (0,11),(11,14),(12,28),(15,29),(25,29):
            copied = org_xliff.copyNodesByTextIndex(s,e)
            xliff = XliffSegment(copied[0].parentNode,formatter)
            result = ''.join(tags)[s:e]
            self.assertEqual(str(xliff), result)

class PasteNodesAtTextIndex(XliffTest):
    def testPasteNodesAtTextIndex(self):
        """Copy nodes by text index"""
        org_str = 'first <tag>in tag</tag> some more text'
        formatter = FullFormatter()
        for s,e,i in (0,12,3),(12,24,11),(15,29,32),(25,29,0),(25,29,38):
            result = org_str[:i] + org_str[s:e] + org_str[i:]
            xliff = XliffSegment(self.segnode.cloneNode(1),formatter,protect_tags = False)
            copied = xliff.copyNodesByTextIndex(s,e)
            xliff.pasteNodesAtTextIndex(i,copied)
            self.assertEqual(str(xliff), result)

##class CopyNodes(XliffTest):
##    def testCopyNodesFullText(self):
##        """Copy nodes"""
##        tags = ['first ','<tag>','in tag','</tag>',' some more text']
##        formatter = FullFormatter()
##        org_xliff = XliffSegment(self.segnode,formatter)
##        for i in range(org_xliff.nodeCount()):
##            copied = org_xliff.copyNodes(0,i)
##            xliff = XliffSegment(copied,formatter)
##            result = ''.join(tags[0:i])
##            self.assertEqual(str(xliff), result)
##
##        for i in range(len(org_xliff.nodelist)):
##            copied = org_xliff.copyNodes(i,i+1)
##            xliff = XliffSegment(copied,formatter)
##            result = ''.join(tags[i:i+1])
##            self.assertEqual(str(xliff), result)
##
##class PasteNodes(XliffTest):
##    def testCopyNodesFullText(self):
##        """Paste nodes"""
##        tags = ['first ','<tag>','in tag','</tag>',' some more text']
##        formatter = FullFormatter()
##        org_xliff = XliffSegment(self.segnode,formatter)
##
##        copied = org_xliff.copyNodes(0,org_xliff.nodeCount())
##        xliff = XliffSegment(self.segnode.cloneNode(True),formatter)
##        r = tags[:]
##        xliff.pasteNodes(0,copied.childNodes)
##        r += tags
##        xliff.pasteNodes(xliff.nodeCount(),copied.childNodes)
##        r += tags
##        xliff.pasteNodes(3,copied.childNodes)
##        r = r[:3]  + tags + r[3:]
##        
##        result = ''.join(r)
##        self.assertEqual(str(xliff), result)
##
##class SplitTextNode(XliffTest):
##    def testSplitTextNode(self):
##        """Split Text node"""
##        tags = ['fir', 'st ','<tag>','in tag','</tag>',' some more text']
##        formatter = FullFormatter()
##        xliff = XliffSegment(self.segnode,formatter)
##        xliff.splitTextNode(0,3)
##        xliff.removeNodes(1,2)
##        result = ''.join(tags[:1] + tags[2:])
##        self.assertEqual(str(xliff), result)
##
##class RemoveNodes(XliffTest):
##    def testRemoveNodes(self):
##        """Remove nodes"""
##        tags = ['first ','<tag>','in tag','</tag>',' some more text']
##        formatter = FullFormatter()
##        xliff = XliffSegment(self.segnode,formatter)
##        xliff.removeNodes(2,4)
##        result = ''.join(tags[:2] + tags[4:])
##        self.assertEqual(str(xliff), result)
##
##class CutNodes(XliffTest):
##    def testCutNodes(self):
##        """Cut nodes"""
##        tags = ['first ','<tag>','in tag','</tag>',' some more text']
##        formatter = FullFormatter()
##        org_xliff = XliffSegment(self.segnode,formatter)
##
##        cut = org_xliff.cutNodes(1,3)
##        result = ''.join(tags[1:3])
##        xliff = XliffSegment(cut,formatter)
##        self.assertEqual(str(xliff), result)
##        result = ''.join(tags[:1] + tags[3:])
##        self.assertEqual(str(org_xliff), result)

class EditTextNode(XliffTest):                          
    def testInsertText(self):
        """Insert text in a text node"""
        result = 'firstsecond <tag>in tag</tag> some more text'
        formatter = FullFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        xliff.insertText(5,'second')
        self.assertEqual(str(xliff), result)

    def testInsertTextAtEnd(self):
        """Insert text in a text node"""
        result = 'first <tag>in tag</tag> some more textend'
        formatter = FullFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        xliff.insertText(38,'end')
        self.assertEqual(str(xliff), result)

##class EditElementText(XliffTest):                          
##    def testInsertText(self):
##        """Insert text in a element node"""
##        result = 'first <modifiedtag>in tag</tag> some more text'
##        formatter = FullFormatter()
##        xliff = XliffSegment(self.segnode,formatter)
##        xliff.insertText(7,'modified')
##        self.assertEqual(str(xliff), result)

##class DeleteElementText(XliffTest):                          
##    def testDeleteText(self):
##        """Delete text in a element node"""
##        result = 'first <ag>in tag</tag> some more text'
##        formatter = FullFormatter()
##        xliff = XliffSegment(self.segnode,formatter)
##        xliff.deleteText(7,8)
##        self.assertEqual(str(xliff), result)
##
##    def testDeleteTextAtEnd(self):
##        """Delete text in a element node"""
##        result = 'first <tag>in tag</tag> some more te'
##        formatter = FullFormatter()
##        xliff = XliffSegment(self.segnode,formatter)
##        xliff.deleteText(36,38)
##        self.assertEqual(str(xliff), result)

class DeleteElementsText(XliffTest):                          
    def testDeleteFullSpanningText(self):
        """Delete text spanning multiple text and element nodes"""
        org = 'first <tag>in tag</tag> some more text'
        for s,e in ((0,2),(1,15),(10,18),(5,20)):
            self.setUp()
            result = org[:s] + org[e:]
            formatter = FullFormatter()
            xliff = XliffSegment(self.segnode,formatter)
            xliff.deleteText(s,e)
            self.assertEqual(str(xliff), result)

    def testDeleteMinimalSpanningText(self):
        """Delete text spanning multiple text and element nodes"""
        org = 'first <>in tag</> some more text'
        for s,e in ((0,2),(1,17),(10,18),(5,20)):
            self.setUp()
            result = org[:s] + org[e:]
            formatter = MinimalFormatter()
            xliff = XliffSegment(self.segnode,formatter)
            xliff.deleteText(s,e)
            self.assertEqual(str(xliff), result)


class EditLockedElementText(XliffTest):                          
    def testInsertText(self):
        """Insert text in a element node should raise error if it's protected"""
        formatter = FullFormatter()
        xliff = XliffSegment(self.segnode,formatter,protect_tags=True)
        self.assertRaises(XliffError, xliff.insertText, 7,'modified')

class XliffBadInput(XliffTest):
    def testNegative(self):
        """Formatter.getNodesInSpan should fail with negative input"""
        formatter = FullFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        self.assertRaises(ValueError, xliff.getNodeByTextIndex, -1)

class CheckFullIndexKnown(XliffTest):
    known_values = {(0,):(0,0),(1,):(0,1),
                    (5,):(0,5),(6,):(1,0),
                    (8,):(1,2),(10,):(1,4),
                    (11,):(2,0),(18,):(3,1),
                    }
##0123456789abcdef0123456789abcdef123456
##first <tag>in tag</tag> some more text

    def testFullFormatter(self):
        """Xliff full formatter should give correct node and sub index"""
        formatter = FullFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        f = xliff.getNodeByTextIndex
        for args,result in self.known_values.items():
            self.assertEqual(apply(f,args), (self.nodelist[result[0]],result[1]))

class LastIndexTest(XliffTest):
##0123456789abcdef0123456789abcdef123456
##first <tag>in tag</tag> some more text

    def testLastIndex(self):
        """getLast() should give the correct last index"""
        formatter = FullFormatter()
        xliff = XliffSegment(self.segnode,formatter)
        result = xliff.getLast()
        self.assertEqual((self.nodelist[-1],15), result)

##class CheckMinimalIndexKnown(XliffTest):
##    known_values = {(0,):(0,0),(1,):(0,1),
##                    (5,):(0,5),(6,):(1,0),
##                    (8,):(2,0),(10,):(2,2),
##                    (15,):(3,1),(18,):(4,1),
##                    }
####0123456789abcdef0123456789abcdef123456
####first <>in tag</> some more text
##
##    def testMinimalFormatter(self):
##        """Xliff using minimal formatter should give correct node and sub index"""
##        formatter = MinimalFormatter()
##        xliff = XliffSegment(self.segnode,formatter)
##        f = xliff.getNodeByTextIndex
##        for args,result in self.known_values.items():
##            self.assertEqual(apply(f,args), result)

##class CheckTextSpanKnown(XliffTest):
##    known_values = {(0,):(0,6),(1,):(6,11),
##                    (2,):(11,17),(3,):(17,23),
##                    (4,):(23,38),
##                    }
####0123456789abcdef0123456789abcdef123456
####first <tag>in tag</tag> some more text
####first <>in tag</> some more text
##
##    def testMinimalFormatter(self):
##        """Full formatter should give correct node span"""
##        formatter = FullFormatter()
##        xliff = XliffSegment(self.segnode,formatter)
##        f = xliff.getTextSpanOfNode
##        for args,result in self.known_values.items():
##            self.assertEqual(apply(f,args), result)


class TestInsertTextUndo(XliffTest):                          
    def testInsertTextUndo(self):
        """Test Insert text undo """
        result = 'firmodifiedst <tag>in tag</tag> some more text'
        result2 = 'first <tag>in tag</tag> some more text'
        formatter = FullFormatter()
        xliff = Model(self.segnode,formatter)
        xliff.insertTextWithUndo(3,'modified')
        self.assertEqual(str(xliff), result)
        xliff.undo()
        self.assertEqual(str(xliff), result2)

class TestDeleteTextRedo(XliffTest):                          
    def testDeleteTextRedo(self):
        """Test Delete text redo """
        result = 'rst <tag>in tag</tag> some more text'
        result2 = 'first <tag>in tag</tag> some more text'
        formatter = FullFormatter()
        xliff = Model(self.segnode,formatter)
        xliff.deleteTextWithUndo(0,2)
        self.assertEqual(str(xliff), result)
        xliff.undo()
        self.assertEqual(str(xliff), result2)
        xliff.redo()
        self.assertEqual(str(xliff), result)

class TestDeleteTextUndo(XliffTest):                          
    def testDeleteTextUndo(self):
        """Test Delete text undo """
        result = 'rst <tag>in tag</tag> some more text'
        result2 = 'first <tag>in tag</tag> some more text'
        formatter = FullFormatter()
        xliff = Model(self.segnode,formatter)
        xliff.deleteTextWithUndo(0,2)
        self.assertEqual(str(xliff), result)
        xliff.undo()
        self.assertEqual(str(xliff), result2)

class TestCutTextUndo(XliffTest):                          
    def testCutTextUndo(self):
        """Test Cut text undo """
        result = 'rst <tag>in tag</tag> some more text'
        result2 = 'first <tag>in tag</tag> some more text'
        formatter = FullFormatter()
        xliff = Model(self.segnode,formatter)
        xliff.cutNodesByTextIndexWithUndo(0,2)
        self.assertEqual(str(xliff), result)
        xliff.undo()
        self.assertEqual(str(xliff), result2)

class TestPasteTextUndo(XliffTest):                        
    def testDeleteTextUndo(self):
        """Test Paste undo """
        org_str = 'first <tag>in tag</tag> some more text'
        formatter = FullFormatter()
        xliff = Model(self.segnode,formatter)

        for s,e,i in (0,12,3),(12,24,11),(15,29,32),(25,29,0),(25,29,38):
            result = org_str[:i] + org_str[s:e] + org_str[i:]
            xliff = Model(self.segnode.cloneNode(1),formatter,protect_tags = False)
            copied = xliff.copyNodesByTextIndex(s,e)
            xliff.pasteNodesAtTextIndexWithUndo(i,copied)
            self.assertEqual(str(xliff), result)
            xliff.undo()
            self.assertEqual(str(xliff),org_str )

##class TestSetFormatterUndo(XliffTest):                          
##    def testCutTextUndo(self):
##        """Test Cut text undo """
##        result = 'rst <tag>in tag</tag> some more text'
##        result2 = 'first <tag>in tag</tag> some more text'
##        full_formatter = FullFormatter()
##        min_formatter = MinimalFormatter()
##        xliff = Model(self.segnode,full_formatter)
##        xliff.setFormatterWithUndo(min_formatter)
##        self.assertEqual(xliff.formatter, min_formatter)
##        xliff.undo()
##        self.assertEqual(xliff.formatter, full_formatter)


if __name__ == "__main__":
    unittest.main()