__doc__ ='''
Main XLIFF inline tag model used to modify the xml directly using char index operations.
Can use different text formatters.
'''

from xml.dom import minidom,IndexSizeErr              
from sets import Set
from formatters import FullFormatter,MinimalFormatter
from general import XliffError
from transolution.observer import Observable
from transolution.xmlutils import getText,getNodeSpan,removeNodes,copyNodes,pasteNodes,splitText
from transolution.xmlutils import getNodeAttr,setNodeAttr

from modelInterface import ModelInterface

class RootNodeEmptyException(XliffError):pass
class RootNodeIsNone(XliffError):pass

class XliffSegment(ModelInterface):
    'wraps a xliff source/target segment with text formatter and text index operations'
    
    def __init__(self,rootnode,formatter,protect_tags = False,locked=False,parent=None):
        self.setRootNode(rootnode)
        self.protect_tags = protect_tags
        self.locked = locked
        self.sub_segs = {}
        self.setFormatter(formatter)

    def getDocument(self,node=None):
        'return nodes document'
        if not node: node = self.rootnode
        return node.ownerDocument

    def setRootNode(self,rootnode):
        self.rootnode = rootnode

    def getRootNode(self):
        return self.rootnode
        
    def setFormatter(self,formatter):
        self.formatter = formatter
    def getFormatter(self):
        return self.formatter

    def getNodeCount(self):
        return len(self.rootnode.childNodes)

    def getNode(self,node):
        if node in self.rootnode.childNodes:
            return node
        else:
            raise XliffError('Node not in segment')
        
    def isLocked(self,n=None):
        'return True if a node is locked, False if not'
        if n is None:
            locked = self.locked
        elif self.isTag(n) and self.protect_tags:
            locked = True
        else: locked = False
        return locked

    def isTag(self,n):
        'return True if a node is locked, False if not'
        node = self.getNode(n)
        if node.nodeType == node.TEXT_NODE:
            tag = False
        else:
##            if node.namespaceURI == 'http://eviltrans.sourceforge.net':
##                return False
            tag = True
        return tag


    def nodeLen(self,node):
        return len(self.node2text(node))
    
    def node2text(self,node):
        r = self.formatter.node2text(node)
        return r
        
    def getNodeByTextIndex(self,str_i):
        '''
        return a tuple (node_index,sub_index_of_that_node)'''
        if str_i < 0:               raise ValueError('Negative text index not supported')
        if self.rootnode is None:   raise RootNodeIsNone('Root node is None')
        # make sure root node is not empty
        self.assureRootNotEmpty()

        chi = 0
        for node in self.rootnode.childNodes:
            old_chi = chi
            chi += self.nodeLen(node)
            # check if index is in node
            if str_i < chi:
                sub_i = str_i - old_chi
                break

        if not str_i < chi:
            raise IndexError('Text index out of range')

        return node,sub_i

    def assureRootNotEmpty(self):
        # make sure root node is not empty as most operations rely on at least one (empty) childnode
        if not self.rootnode.hasChildNodes():
            self.rootnode.appendChild(self.rootnode.ownerDocument.createTextNode(''))

    def getTextNodeByTextIndex(self,str_i):
        '''
        return a tuple (node_index,sub_index_of_that_node),
        returns the prev text node or create one if node at index is a tag and offset is 0.
        '''
        self.assureRootNotEmpty()

        try:node,offset = self.getNodeByTextIndex(str_i)
        except IndexError,e:
            # HACK: if index is beond end return the last index
            node,offset = self.getLast()
            if self.isTag(node) and offset == self.nodeLen(node):
                # HACK: if node is the last tag and offset is at end create a text node at the end instead
                node = self.rootnode.appendChild(self.getDocument().createTextNode(''))
                offset = 0
        else:
            if self.isTag(node) and offset == 0:
                # HACK: if node is a tag and offset is 0 return prev textnode last index or create a text node instead
                prev = node.previousSibling
                if not prev or self.isTag(prev):
                    node = self.rootnode.insertBefore(self.getDocument().createTextNode(''),node)
                else:
                    node = prev
                offset = self.nodeLen(node)
        return node,offset
    
    def getTextSpanOfNode(self,node):
        '''
        return a tuple (start_text_index,end_text_index)'''
        self.assureRootNotEmpty()

        chi = 0
        for n in self.rootnode.childNodes:
            t = self.node2text(n)
            old_chi = chi

            chi += len(t)
            if node is n:
                return (old_chi,chi)

        raise XliffError('node not found in nodelist:%s'%(node,))

    
    def pasteNodesAtTextIndex(self,ti,nodelist):
        'copy nodes in range'
##        try:
        node,offset = self.getTextNodeByTextIndex(ti)
        n1,n2 = self._split(node,offset)
        pasteNodes(n2,nodelist)
##        except RootNodeEmptyException,e:
##            # if it its empty just copy the nodes down
##            for c in nodelist:
##                self.rootnode.appendChild(c.cloneNode(True))
        
        
    def cutNodesByTextIndex(self,start_ti,end_ti):
        'cut nodes in range'
        ret = self.copyNodesByTextIndex(start_ti,end_ti)
        self.deleteText(start_ti,end_ti)
        return ret


    def copyNodesByTextIndex(self,start_ti,end_ti):
        node,offset = self.getTextNodeByTextIndex(start_ti)
        n1,n2 = self._split(node,offset)
        node,offset = self.getTextNodeByTextIndex(end_ti)
        n3,n4 = self._split(node,offset)
##        for node in getNodeSpan(n2,n4):
##            if self.isLocked(node): raise XliffError('Node protected')
        return copyNodes(n2,n4)


    def insertText(self,ti,text):
        if self.isLocked():
            raise XliffError('Model protected')
        node,offset = self.getTextNodeByTextIndex(ti)
        if self.isTag(node):
            raise XliffError('Node is tag')
        else:
            try:
                node.insertData(offset,text)
            except IndexSizeErr,e:
                node.appendData(text)


    def deleteText(self,start_ti,end_ti,force=False):
        'remove nodes in span, return a copy of the removed nodes to allow undo'
        if self.isLocked():
            raise XliffError('Model protected')

        node,offset = self.getTextNodeByTextIndex(start_ti)
        n1,n2 = self._split(node,offset)
        node,offset = self.getTextNodeByTextIndex(end_ti)
        n3,n4 = self._split(node,offset)
        if not force:
            for node in getNodeSpan(n2,n4):
                if self.isLocked(node): raise XliffError('Node protected')

        ret = copyNodes(n2,n4)
        removeNodes(n2,n4)
        self.rootnode.normalize()
        #HACK: ugly hack to make sure there always is a end text node, needed to make deleting at end work
        self.rootnode.appendChild(self.getDocument().createTextNode(''))

        return ret

    def clear(self,interactive=True):
        try:
            self.deleteText(0,len(self),force=True)
        except RootNodeEmptyException,e:
            pass

    def _split(self,node,str_i):
        'split node into nodes at text index'
        if self.isLocked(node):
            raise XliffError('Node protected')
        if node.nodeType != node.TEXT_NODE:
            raise XliffError('Node protected:%s'%node)
        else:
##            new_node = node.splitText(str_i)
            # cant use node.splitText as it is minidom specific :-( so our own implementation
            new_node = splitText(node,str_i)
        return new_node.previousSibling,new_node


    def getLast(self):
        '''
        return the last index. returns last node and its length as a tuple
        '''
        node = self.rootnode.lastChild
        
        if node is None:
            raise RootNodeEmptyException('Root node empty')
##            return node,0
        sub_i =self.nodeLen(node)
        return node,sub_i

    def getTagNodeByNameAndID(self,tagname,id):
        '''
        return firs child node fullfilling the creteria
        '''
        for n in self.rootnode.childNodes:
            if n.nodeType == n.ELEMENT_NODE:
                if n.tagName == tagname:
                    try:
                        if getNodeAttr(n,'id') == id: return n
                    except KeyError,e:
                        continue
        raise XliffError('No node matching creteria found in nodelist')

    def getTagNodeByID(self,id):
        '''
        yield child nodes fulfilling the creteria
        '''
        for n in self.rootnode.childNodes:
            if n.nodeType == n.ELEMENT_NODE:
                try:
                    if getNodeAttr(n,'id') == id:
                        yield n
                except KeyError,e:
                    continue

    def getHighestId(self):
        'return highest tag id number used'
        id = 0
        for n in self.rootnode.childNodes:
            if n.nodeType == n.ELEMENT_NODE:
                try: id = max(int(getNodeAttr(n,'id')),id)
                except KeyError,e:
                    continue
        return id
        
    def nodelist2string(self,nodelist):
        ret = []
        for node in nodelist:
            ret.append(self.node2text(node))
        return ''.join(ret)
        
    def __str__(self):
        return self.nodelist2string(self.rootnode.childNodes)
    
    def __len__(self):
        return len(str(self))

    def getFormatted(self):
        'get a list of tupels with the (style,text)'
        ret = []
        for node in self.rootnode.childNodes:
            ret.append(self.formatter.getFormatted(node))
        return ret


from undoredo import UndoRedo,State
class UndoRedoXliff(XliffSegment,UndoRedo):
    'Adds undo redo functions to Xliff'
    def __init__(self,*args,**kwargs):
        self.rootnode = None
        XliffSegment.__init__(self,*args,**kwargs)
        UndoRedo.__init__(self)
    def setRootNode(self,rootnode):
        'erase the undo redo buffers if rootnode changes'
        if self.rootnode != rootnode:
            self.rootnode = rootnode
            self.undobuffer = []
            self.redobuffer = []

    def clearWithUndo(self,interactive=True):
        try:
            self.deleteTextWithUndo(0,len(self),interactive=interactive,force=True)
        except RootNodeEmptyException,e:
            pass
        
    def getState(self):
        sub_states = []
        return State(formatter = self.formatter,protect_tags = self.protect_tags)

    def setState(self,state):
        for key,val in state.dict.items():
            setattr(self,key,val)
    
        
class Model(UndoRedoXliff,Observable):
    'Adds obeserver pattern interface to Xliff'
    
    def __init__(self,*args,**kwargs):
        UndoRedoXliff.__init__(self,*args,**kwargs)

        # observer pattern interface
        Observable.__init__(self)


if __name__ == "__main__":
    xml = r'''<seg>first <bpt>&lt;tag1<sub><bpt>"</bpt>test<ept>"</ept></sub>&gt;</bpt>in tag<ept>&lt;/tag&gt;</ept> some more text</seg>'''

##    xml = r'''<seg>first <bpt>&lt;tag&gt;</bpt>in tag<ept>&lt;/tag&gt;</ept> some more text</seg>'''
    xmldoc = minidom.parseString(xml)
    formatter = FullFormatter()
    #formatter = MinimalFormatter()
    xliff = XliffSegment(xmldoc.firstChild,formatter)
    xliff.insertText(22,"Test")
    xmlstr = str(xliff)
    print xliff
    
    for i in range(len(str(xliff))):
        node,offset = xliff.getNodeByTextIndex(i)
        si,ei = xliff.getTextSpanOfNode(node)
#        print node
        print i,xliff.isLocked(node),(si,ei),'"%s"'%(xmlstr[si:ei],), xmlstr[i]
    print xmldoc.toxml()


    print xliff
    print xliff.getDocument()
